#
# Virtual Traces Experiment Software Distribution.
# This code is licensed under GPL v2, and is copyright (C) Jack Whitham 2008.
# See PATCH_NOTE.
#


import math

STA = "STA "
TICK = "global: X2 Tick"
COMMIT = "system.cpu.commit: Committed pc"

class BB_Stats:
    def __init__(self, pc):
        self.start = pc
        self.timings = dict()

    def Transition(self, next_bb, ticks):
        try:
            self.timings[ ticks ] += 1
        except KeyError:
            self.timings[ ticks ] = 1

    def Get_Stats(self):
        if ( len(self.timings) == 0 ):
            return None

        total_value = 0
        total_count = 0

        timings_list = self.timings.items()
        timings_list.sort()
        for (ticks, count) in timings_list:
            total_count += count
            total_value += ticks * count

        assert total_count > 0
        mean = float(total_value) / float(total_count)
        total_variance = 0.0

        for (ticks, count) in timings_list:
            v = float(ticks) - mean
            total_variance += v * v * float(count)

        variance = total_variance / float(total_count)
        std_dev = math.sqrt(variance)

        return (total_count, mean, std_dev)

    def Get_Stats_Text(self):
        rc = self.Get_Stats()
        if ( rc == None ):
            return None

        (total_count, mean, std_dev) = rc
        return '0x%010x, %5u, %5.2f, %5.2f' % (
                    self.start, total_count, mean, std_dev)

def Load_Stats(meta_in, commit_out):
    bb_tbl = dict()
    bb_list = []

    for line in file(meta_in):
        if ( line.startswith(STA) ):
            pc = int(line.split()[ 1 ], 16)
            bb_tbl[ pc ] = bb = BB_Stats(pc)
            bb_list.append(bb)

    start_tick = ticks = 0
    bb = None
    for line in file(commit_out):
        x = line.find(':')
        if ( x < 0 ):
            continue

        x += 2
        line = line[ x: ]
        if ( line.startswith(TICK) ):
            ticks += 1

        elif ( line.startswith(COMMIT) ):
            pc = int(line[ len(COMMIT): ].split()[ 0 ], 16)

            if ( bb_tbl.has_key(pc) ):
                # Basic block boundary
                next_bb = bb_tbl[ pc ]
                    
                if ( bb != None ):
                    bb.Transition(next_bb, ticks - start_tick)

                bb = next_bb
                start_tick = ticks

    return bb_list

def Main():
    stats = Load_Stats("matmult_meta.txt", "matmult_run_01")
    out = file("matmult_stats.txt", 'wt')
    for bb in stats:
        rc = bb.Get_Stats_Text()
        if ( rc != None ):
            out.write(rc)
            out.write('\n')
    out.close()


