
import sys, os

[ BT_None, BT_Unconditional, BT_Conditional, BT_Serialising ] = "NUCS"
BT_LOOKUP = {
        "None" : BT_None,
        "Unconditional" : BT_Unconditional,
        "Conditional" : BT_Conditional,
        "Serialising" : BT_Serialising,
        }

class CFG_Vertex:
    def __init__(self, id):
        self.id = id
        self.pc = None
        self.entrances_pass_one = []
        self.exits_pass_one = []
        self.entrances = None
        self.exits = None
        self.udg_printed = False
        self.identify_loops = False
        self.is_exit = False

    def Import_Data(self, fields, table):
        key = fields[ 0 ]
        if ( key == "entrance" ):
            table[ key ] = self
            return
        elif ( key == "exit" ):
            table[ key ] = self
            self.is_exit = True
            return

        value = int(fields[ 1 ], 16)
        if ( key == "pc" ):
            self.pc = value
        elif ( key == "ie" ):
            self.entrances_pass_one.append(value)
        elif ( key == "oe" ):
            self.exits_pass_one.append(value)

    def Pass_Two(self, table):
        self.entrances = []
        self.exits = []
        for i in self.entrances_pass_one:
            self.entrances.append(table[ i ])
        for i in self.exits_pass_one:
            self.exits.append(table[ i ])
        self.entrances_pass_one = None
        self.exits_pass_one = None

    def Clear_UDG_Flag(self):
        self.udg_printed = False

    def Export_UDG(self, udg_out, edge_label_fn, 
                    vertex_label_fn=None, root=True):
        if self.udg_printed:
            if root:
                return False
            else:
                udg_out.write(r'r("L%x")' % self.id)
                return True

        self.udg_printed = True
        vertex_label = ""
        if vertex_label_fn != None:
            vertex_label = vertex_label_fn(self)

        udg_out.write('l("L%x",n("C%x",[a("OBJECT","%s")],[' % (self.id, 
                self.id, vertex_label))

        first = True
        for e in self.exits:
            if not first:
                udg_out.write(',')

            first = False
            udg_out.write('e("E%x",[a("OBJECT","")],'
                'l("Z%x",n("Y%x",[a("OBJECT","%s")],['
                'e("X%x",[a("OBJECT","")],' % (e.id, e.id, e.id,
                        edge_label_fn(e), e.id))
            e.end.Export_UDG(udg_out, edge_label_fn, vertex_label_fn, False)
            udg_out.write(')])))')

        udg_out.write(']))')
        return True


class CFG_Edge:
    def __init__(self, id):
        self.id = id
        self.frequency = self.start = self.end = None
        self.path = []
        self.branch_type = BT_None
        self.superblock_next = None
        self.transition_limit = 10000
        self.superblock = None
        self.cost = 0
        self.max_frequency = None
        self.pc = None
        self.end_pc = None
        self.loop_body = None
        self.loop_header = None
        self.is_dirty = False

    def Import_Data(self, fields, table):
        key = fields[ 0 ]
        try:
            value = int(fields[ 1 ], 16)
        except:
            value = fields[ 1 ]

        if ( key == "f" ):
            self.frequency = value
            self.max_frequency = value
        elif ( key == "iv" ):
            self.start = table[ value ]
        elif ( key == "ov" ):
            self.end = table[ value ]
        elif ( key == "x" ):
            if ( self.pc == None ):
                self.pc = value
            self.path.append(value)
            self.end_pc = value + 4
        elif ( key == "cost" ):
            self.cost = value
        elif ( key == "bt" ):
            self.branch_type = BT_LOOKUP[ value ]

def Find_V_E(entrance):
    assert isinstance(entrance, CFG_Vertex)

    out_vertexes = []
    out_edges = []
    seen_before = set()
    todo = [ entrance ]

    while ( len(todo) != 0 ):
        element = todo.pop()
        if ( element.id in seen_before ):
            continue
        
        seen_before.add(element.id)

        if ( isinstance(element, CFG_Vertex) ):
            out_vertexes.append(element)
            for edges in (element.entrances, element.exits):
                for edge in edges:
                    todo.append(edge)

        elif ( isinstance(element, CFG_Edge) ):
            out_edges.append(element)
            if ( element.start != None ):
                todo.append(element.start)

            if ( element.end != None ):
                todo.append(element.end)

        else:
            assert False, "CFG element is not Edge or Vertex"

    return (out_vertexes, out_edges)
    

def Read_CFG(fname, vertex_class=CFG_Vertex, edge_class=CFG_Edge):
    assert issubclass(vertex_class, CFG_Vertex)
    assert issubclass(edge_class, CFG_Edge)
    vertex = edge = None
    vlist = []
    table = dict()
    for line in file(fname):
        fields = line.split()

        if (( line[ 1 ] != ' ' )
        or ( not ( line[ 0 ] in " ve" ))):
            vertex = edge = None

        elif ( line[ 0 ] == 'v' ):
            edge = None
            id = int(fields[ 1 ], 16)
            vertex = vertex_class(id)
            table[ id ] = vertex
            vlist.append(vertex)

        elif ( line[ 0 ] == 'e' ):
            vertex = None
            id = int(fields[ 1 ], 16)
            edge = edge_class(id)
            table[ id ] = edge

        else:
            if ( vertex != None ):
                vertex.Import_Data(fields, table)
            elif ( edge != None ):
                edge.Import_Data(fields, table)

    for vertex in vlist:
        vertex.Pass_Two(table)
    
    return table[ "entrance" ]

def Identify_Loops(entrance):
    assert isinstance(entrance, CFG_Vertex)
    loop_list = []
    edge_history = []
 
    def Vertex(v1):
        if len(v1.entrances) > 1:
            # Join point
            assert len(v1.exits) == 1
            loop_header = None
            for e in edge_history[ : -1 ]:
                if v1 == e.end:
                    loop_header = e
                    break

            if loop_header != None:
                loop_header.loop_body = v1.exits[ 0 ]
                loop_header.loop_body.loop_header = loop_header
                loop_list.append(loop_header)
        else:
            # Not a join point
            pass

        if v1.identify_loops:
            return
        else:
            v1.identify_loops = True 

        for e1 in v1.exits:
            if e1.loop_header == None:
                edge_history.append(e1)
                Vertex(e1.end)
                e2 = edge_history.pop()
                assert e2 == e1


    Vertex(entrance)
    return loop_list



