#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#define M5_CMD              "m5.opt"

#include "makecfg.hh"
#include "unix.hh"

Symbol_Table * Do_Read_Symbols ( 
                std::string program , std::string nm_bin )
{
    char            cmd [ BUFSIZ ] ;
    Symbol_Table *  symtab = 0L ;

    cmd [ BUFSIZ - 1 ] = '\0' ;
    snprintf ( cmd , BUFSIZ - 1 ,
            "%s -S %s" , nm_bin . c_str () , program . c_str () ) ;
    fprintf ( stderr , "Obtain symbols: %s\n" , cmd ) ;

    FILE * syfd = popen ( cmd , "r" ) ;
    if ( syfd == 0L )
    {
        perror ( "Unable to obtain symbols." ) ;
        return 0L ;
    }

    symtab = new Symbol_Table ( syfd ) ;
    fclose ( syfd ) ;
    wait ( 0L ) ;
    return symtab ;
}

bool Do_Make_CFG ( std::string program ,
                std::string output ,
                std::string udg_output ,
                std::string nm_bin ,
                std::string m5_path ,
                FILE * exe_trace_out )
{
    char            cmd [ BUFSIZ ] ;
    char            fifo_name [ BUFSIZ ] ;
    int             uid = getpid () ;
    Make_CFG *      m = new Make_CFG () ;
    Symbol_Table *  symtab = 0L ;
    FILE *          m5fd = 0L ;
    FILE *          fifo_fd = 0L ;
    FILE *          out_fd = 0L ;
    ulongint        main_end , main_start , main_size ;
    bool            rc = false ;
    bool            src ;

    cmd [ BUFSIZ - 1 ] = '\0' ;
    fifo_name [ BUFSIZ - 1 ] = '\0' ;
    snprintf ( fifo_name , BUFSIZ - 1 , "/tmp/%d.fifo" , uid ) ;

    fprintf ( stderr , "FIFO name: %s\n" , fifo_name ) ;

    do {
        symtab = Do_Read_Symbols ( program , nm_bin ) ;
        if ( symtab == 0L )
        {
            break ;
        }

        src = symtab -> Get_Record ( "main" , main_start , main_size ) ;
        main_end = main_start + main_size ;

        if ( ! src )
        {
            fprintf ( stderr , "Error reading symbols (main not found)\n" ) ;
            break ;
        }
        
        if ( mkfifo ( fifo_name , 0600 ) != 0 )
        {
            perror ( "Error creating FIFO" ) ;
            break ;
        }

        snprintf ( cmd , BUFSIZ - 1 ,
                "%s/" M5_CMD " --trace-flags=Exec,ExecFTR "
                "--remote-gdb-port=0 "
                "--trace-file %s %s/../../configs/example/se.py "
                "-c %s" , m5_path . c_str () , 
                    fifo_name , m5_path . c_str () ,
                    program . c_str () ) ;
        fprintf ( stderr , "M5: %s\n" , cmd ) ;

        m5fd = popen ( cmd , "r" ) ;
        if ( m5fd == 0L )
        {
            perror ( "Unable to start M5." ) ;
            break ;
        }

        fifo_fd = fopen ( fifo_name , "rb" ) ;
        if ( fifo_fd == 0L )
        {
            perror ( "Unable to open FIFO." ) ;
            break ;
        }


        m -> Read_Trace ( fifo_fd , main_start , main_end , exe_trace_out ) ;
        fclose ( m5fd ) ;
        fclose ( fifo_fd ) ;
        m5fd = fifo_fd = 0L ;
        wait ( 0L ) ;

        fprintf ( stderr , "Making CFG: %s\n" , output . c_str () ) ;
        m -> Make () ;
        out_fd = fopen ( output . c_str () , "wt" ) ;
        if ( out_fd == 0L )
        {
            perror ( "Unable to open output file." ) ;
            break ;
        }

        m -> Export ( out_fd ) ;
        fclose ( out_fd ) ;
        out_fd = 0L ;

        if ( udg_output . size () != 0 )
        {
            out_fd = fopen ( udg_output . c_str () , "wt" ) ;
            if ( out_fd == 0L )
            {
                perror ( "Unable to open UDG output file." ) ;
                break ;
            }
            m -> Export_UDG ( out_fd , symtab ) ;
        }
        rc = true ;
    } while ( 0 ) ;

    delete m ;
    if ( symtab != 0L ) delete symtab ;
    if ( m5fd != 0L ) fclose ( m5fd ) ;
    if ( out_fd != 0L ) fclose ( out_fd ) ;
    if ( fifo_fd != 0L ) fclose ( fifo_fd ) ;

    unlink ( fifo_name ) ;

    return rc ; 
}
